#include "ACO3D.h"
#include "HPmodel.h"
#include <algorithm>
#include <cmath>
#include <iostream>
#include <numeric>
#include <random>

ACO3D::ACO3D(const std::string &sequence, const ACOParams &params)
    : hpSequence(sequence), acoParams(params) {
    n = static_cast<int>(hpSequence.size());
    rng.seed(std::random_device{}());

    pheromone.resize(n - 1, std::vector(5, params.pheromoneInit));
}

char directionToChar(DIRECTION dir) {
    switch(dir) {
        case FORWARD: return 'f';
        case LEFT: return 'l';
        case RIGHT: return 'r';
        case UP: return 'u';
        case DOWN: return 'd';
        default: throw std::invalid_argument("Invalid direction");
    }
}

BestSolution ACO3D::solve() {
    initPheromones();

    std::vector ants(acoParams.numAnts, Ant{});

    BestSolution bestSol{HPModel(hpSequence), "", 1, false, };

    for (int iteration = 0; iteration < acoParams.maxIterations; ++iteration) {
        constructSolutions(ants);

        localSearch(ants);

        for (auto &ant : ants) {
            if (ant.valid && ant.score > bestSol.score) {
                bestSol = {ant.model, convertFoldsToString(ant.folds), ant.score, true};
            }
        }

        updatePheromones(ants);

        if (iteration % 5 == 0 || iteration == acoParams.maxIterations - 1) {
            std::cout << "[Iter " << iteration << "] Best Score = " << bestSol.score << std::endl;
        }
    }

    return bestSol;
}

std::string ACO3D::convertFoldsToString(const std::vector<DIRECTION> &folds) {
    std::string foldString;
    for (const auto &fold : folds) {
        foldString += directionToChar(fold);
    }
    return foldString;
}

void ACO3D::initPheromones() {
    for (auto &row : pheromone) {
        std::fill(row.begin(), row.end(), acoParams.pheromoneInit);
    }
}

void ACO3D::constructSolutions(std::vector<Ant> &ants) {
    for (auto &ant : ants) {
        ant.model = HPModel(hpSequence);
        ant.folds.clear();
        ant.valid = true;

        try {
            // Choose a random index to start folding
            int startIndex = std::uniform_int_distribution(0, n - 2)(rng);

            // Set starting positions for left and right
            int leftPos = startIndex - 1;  // Start from the left of the middle
            int rightPos = startIndex + 1; // Start from the right of the middle

            // Build the chain
            buildChain(ant, leftPos, rightPos);

            if (ant.valid) evaluateAnt(ant);
        } catch (...) {
            ant.valid = false;
        }
    }
}

void ACO3D::buildChain(Ant &ant, int leftPos, int rightPos) {
    while (leftPos >= 0 || rightPos < n-1) {
        // Continue until all residues are folded
        // Check bounds before folding
        if (leftPos < 0 && rightPos >= n-1) {
            break;
        }

        // Calculate the number of unfolded residues on each side
        int unfoldedLeft = std::max(0, leftPos + 1);
        int unfoldedRight = std::max(0, n - rightPos - 1);

        // Decide probabilistically which side to fold based on the ratio of unfolded resiudes
        bool foldLeft = false;
        if (unfoldedLeft > 0 && unfoldedRight > 0) {
            double leftProb = static_cast<double>(unfoldedLeft) / (unfoldedLeft + unfoldedRight);
            foldLeft = std::uniform_real_distribution(0.0, 1.0)(rng) < leftProb;
        } else if (unfoldedLeft > 0) {
            foldLeft = true;
        } else if (unfoldedRight > 0) {
            foldLeft = false;
        }

        // Determine the current position to fold
        int pos = foldLeft ? leftPos : rightPos;

        // Choose a direction for the fold
        DIRECTION dir = chooseDirection(pos, ant);

        // Update sequence and folds
        if (foldLeft) {
            ant.folds.insert(ant.folds.begin(), dir);  // Prepend fold for the left side
            leftPos--;  // Move left index
        } else {
            ant.folds.push_back(dir);  // Append fold for the right side
            rightPos++;  // Move right index
        }

        // Convert folds to a fold string
        std::string foldString;
        for (const auto &fold : ant.folds) {
            foldString += directionToChar(fold);
        }

        // Update the model with the current fold sequence
        try {
            ant.model = HPModel(hpSequence, foldString);
        } catch (...) {
            std::cerr << "Failed to initialize HPModel: " << std::endl;

            // Backtrack if initialization fails
            if (!partialBacktrack(ant, pos)) {
                ant.valid = false;
                return;  // Exit if backtracking fails
            }
        }
    }
}

DIRECTION ACO3D::chooseDirection(int pos, Ant &ant) {
    std::vector probabilities(5, 0.0);
    double totalProbability = 0.0;

    // Compute probabilities for each direction
    for (int d = 0; d < 5; ++d) {
        DIRECTION dir = static_cast<DIRECTION>(d);
        HPModel tempModel = ant.model.clone();
        bool isValid = tempModel.addPoint(dir, ant.model.isCurrentAcidHydrophobic());

        if (!isValid) {
            probabilities[d] = 0.0;
            continue;
        }

        size_t newContacts = tempModel.countNewContactsForHead();
        double eta = std::exp(-acoParams.gamma * static_cast<double>(newContacts)); // Boltzmann heuristic

        // Calculate combined probability
        double tau = pheromone[pos][d];
        probabilities[d] = std::pow(tau, acoParams.alpha) * std::pow(eta, acoParams.beta);
        totalProbability += probabilities[d];
    }

    // Normalize probabilities
    if (totalProbability > 0.0) {
        for (double &prob : probabilities) {
            prob /= totalProbability;
        }
    } else {
        // If total probability is 0, choose uniformly at random
        std::fill(probabilities.begin(), probabilities.end(), 1.0 / 5.0);
    }

    // Create a discrete distribution based on the computed probabilities
    std::discrete_distribution dist(probabilities.begin(), probabilities.end());

    // Select a direction
    int selectedDirection = dist(rng);

    return static_cast<DIRECTION>(selectedDirection);
}


bool ACO3D::tryFold(Ant &ant, DIRECTION dir) {
    try {
        ant.model.fold(dir);
        ant.folds.push_back(dir);
        return true;
    } catch (...) {
        return false;
    }
}

bool ACO3D::partialBacktrack(Ant &ant, int currentFoldIndex) {
    if(currentFoldIndex == 1) {
        throw std::runtime_error("Cannot backtrack further");
    }
    // Simplified backtracking logic
    int rollbackSteps = (currentFoldIndex + 1) / 2;
    ant.folds.resize(currentFoldIndex + 1 - rollbackSteps);
    ant.model = HPModel(hpSequence);

    for (auto dir : ant.folds) {
        if (!tryFold(ant, dir)) return false;
    }

    return true;
}

void ACO3D::localSearch(std::vector<Ant> &ants) {
    int topAnts = std::max(1, acoParams.numAnts / 10);

    for (int i = 0; i < topAnts; ++i) {
        Ant &ant = ants[i];
        if (!ant.valid) continue;

        for (int attempt = 0; attempt < acoParams.localSearchMaxNoImprove; ++attempt) {
            if (!performLongRangeMutation(ant)) break;
            evaluateAnt(ant);
        }
    }
}

bool ACO3D::performLongRangeMutation(Ant &ant) {
    std::uniform_int_distribution<> dist(0, ant.folds.size() - 1);
    int mutationPoint = dist(rng);

    // Mutate the selected point
    std::vector<DIRECTION> directions = {FORWARD, LEFT, RIGHT, UP, DOWN};
    std::shuffle(directions.begin(), directions.end(), rng);

    for (auto dir : directions) {
        if (dir != ant.folds[mutationPoint]) {
            // Temporarily mutate
            auto originalDir = ant.folds[mutationPoint];
            ant.folds[mutationPoint] = dir;

            // Adapt the rest of the chain probabilistically
            for (size_t i = mutationPoint + 1; i < ant.folds.size(); ++i) {
                std::vector<double> probabilities(5, 0.0);
                double totalProbability = 0.0;

                for (int d = 0; d < 5; ++d) {
                    DIRECTION newDir = static_cast<DIRECTION>(d);
                    HPModel tempModel = ant.model.clone();
                    if (!tempModel.addPoint(newDir, ant.model.isCurrentAcidHydrophobic())) continue;

                    double eta = std::exp(-acoParams.gamma * tempModel.countNewContactsForHead());
                    probabilities[d] = std::pow(pheromone[i][d], acoParams.alpha) * std::pow(eta, acoParams.beta);
                    totalProbability += probabilities[d];
                }

                if (totalProbability > 0.0) {
                    for (auto &prob : probabilities) prob /= totalProbability;
                    std::discrete_distribution<> dist(probabilities.begin(), probabilities.end());
                    ant.folds[i] = static_cast<DIRECTION>(dist(rng));
                } else {
                    ant.folds[i] = directions[rng() % directions.size()]; // Fallback random direction
                }
            }

            // Evaluate the new conformation
            // Creating new hp model with the folds
            HPModel newModel = HPModel(hpSequence, convertFoldsToString(ant.folds));
            double newScore = newModel.getScore();
            if (newScore > ant.score) {
                ant.score = newScore;
                return true; // Accept the mutation
            } else {
                ant.folds[mutationPoint] = originalDir; // Revert if no improvement
            }
        }
    }

    return false; // Mutation unsuccessful
}

void ACO3D::updatePheromones(const std::vector<Ant> &ants) {
    // Step 1: Evaporation
    for (size_t i = 0; i < pheromone.size(); ++i) {
        for (size_t j = 0; j < pheromone[i].size(); ++j) {
            pheromone[i][j] *= (1.0 - acoParams.rho);
        }
    }

    // Step 2: Reinforcement
    for (const auto &ant : ants) {
        if (!ant.valid) continue;

        for (size_t pos = 0; pos < ant.folds.size(); ++pos) {
            pheromone[pos][ant.folds[pos]] += ant.score;
        }
    }

    // Step 3: Thresholding and Normalization
    for (size_t i = 0; i < pheromone.size(); ++i) {
        double sumTau = std::accumulate(pheromone[i].begin(), pheromone[i].end(), 0.0);

        if (sumTau == 0.0) continue;

        double minTau = *std::min_element(pheromone[i].begin(), pheromone[i].end());
        if (minTau < acoParams.theta) {
            double delta = acoParams.theta - minTau;

            for (double &tau : pheromone[i]) {
                tau += delta;
            }

            double maxTau = *std::max_element(pheromone[i].begin(), pheromone[i].end());
            for (double &tau : pheromone[i]) {
                tau = (tau - delta) / (maxTau - delta);
            }
        } else {
            for (double &tau : pheromone[i]) {
                tau /= sumTau; // Normalize
            }
        }
    }
}

void ACO3D::evaluateAnt(Ant &ant) {
    ant.score = ant.model.getScore();
}