#pragma once
#include <vector>
#include <string>
#include <random>
#include <map>
#include <set>
#include "HPModel.h"

/**
 * A small struct to store general ACO parameters
 * that parallel the classical 2D HP model approach.
 */
struct ACOParams {
    int numAnts;            ///< number of ants
    int maxIterations;      ///< maximum ACO iterations
    double alpha;           ///< exponent for pheromone
    double beta;            ///< exponent for heuristic
    double rho;             ///< pheromone evaporation factor
    double pheromoneInit;   ///< initial pheromone value
    double stagnationThreshold; ///< threshold for re-normalization
    double theta;
    double mutationRetentionProb;

    // Additional parameters for local search
    int macroMutationAttempts;   ///< how many macro-mutations per iteration
    int pointMutationAttempts;   ///< how many 1-exchange steps per iteration
    int localSearchMaxNoImprove; ///< stop local search after these many consecutive no-improves

     // Add gamma for the heuristic exponent
    double gamma;
};

/**
 * Each ant has:
 *  - A conformation (HPModel).
 *  - A list of directions used so far (length n-1 for an n-residue protein).
 *  - A boolean flag 'valid' to indicate if it constructed a feasible solution.
 *  - The final energy/score from HPModel::getScore().
 */
struct Ant {
 HPModel model;
 std::vector<DIRECTION> folds; // Should have exactly n - 2 elements
 int score;
 bool valid;
 int startFold; // Add to track the starting fold index
};
/**
 * A small structure to store the best solution found so far.
 */
struct BestSolution {
    HPModel model;
    std::string folds;
    int score;
    bool valid;
};

/**
 * A class implementing the 3D HP-model ACO as described in the papers,
 * adapted from the 2D version.
 */
class ACO3D {
public:
    /**
     * Constructor
     * @param sequence The HP sequence (string of 'h' and 'p').
     * @param params   The ACO parameters (see above).
     */
    ACO3D(const std::string &sequence, const ACOParams &params);

    /**
     * Main solver loop (like "run") that:
     *  - Initializes pheromones and heuristics
     *  - Repeatedly calls constructSolutions, localSearch, and updatePheromones
     *  - Tracks the best solution
     */
    BestSolution solve();

private:
    /// The HP sequence we want to fold
    std::string hpSequence;

    /// ACO parameters
    ACOParams acoParams;

    /// length of the protein
    int n;

    /**
     * Pheromone and heuristic storage.
     * We store pheromone[i][d], heuristic[i][d] for i in [0..(n-2)] and d in {REL_S, REL_L, REL_R, REL_U, REL_D}.
     * Why (n-1)? Because for an n-residue protein, we can have up to (n-1) folds.
     * If you want to “fix” the first two residues to remove rotational redundancy,
     * you can treat the first fold as fixed, and only store pheromones from i=1..(n-2).
     */
    std::vector< std::vector<double> > pheromone;

    /// Random number generator
    std::mt19937 rng;

private:
    /** ================== ACO SUB-STEPS ================== **/

    /**
     * Initialize all pheromones to a constant (params.pheromoneInit).
     * Initialize the heuristic table, for instance using the idea from the papers:
     * \eta_{i,d} = # of potential new H-H contacts or exp(...) with an “inverse temperature” factor.
     */
    void initPheromones();

    /**
     * For each ant, build a feasible 3D conformation (or mark it invalid).
     * This uses the classical ACO probability rule:
     *
     *  p_{i,d} = [tau_{i,d}]^alpha * [eta_{i,d}]^beta
     *            -----------------------------------
     *            sum_{e in {S,L,R,U,D}} [tau_{i,e}]^alpha * [eta_{i,e}]^beta
     *
     * Then tries to fold the residue i with direction d. If infeasible, tries next direction,
     * or triggers partial backtracking if no direction can be placed.
     */
    void constructSolutions(std::vector<Ant> &ants);

    /**
     * Attempt to extend the chain by folding with direction 'dirIndex' at fold index i
     * Catch HPC model exceptions if not feasible.
     * Return true if successful, false if infeasible.
     */
    bool tryFold(Ant &ant, DIRECTION dirIndex);

    /**
     * Partial backtracking if the chain cannot extend beyond a certain residue
     * (i.e., all directions are blocked).
     * We backtrack half the distance already folded and attempt a different direction for that residue,
     * then re-fold forward from there (the approach from the 2D paper).
     * Return false if we cannot recover from this block (the ant becomes invalid).
     */
    bool partialBacktrack(Ant &ant, int currentFoldIndex);

    /**
     * Local search that follows the approach from the papers:
     *   1) Macro-mutation neighborhood: randomly pick 2 positions i < j and re-randomize the folds in [i..j].
     *   2) 1-exchange: for each residue’s fold, try changing it if it improves the energy.
     *
     * We do an iterative approach:
     *   - Macro-mutation step (random block re-fold).
     *   - 1-exchange steps in random order.
     *   - Stop if no improvements or we exceed localSearchMaxNoImprove.
     */
    void localSearch(std::vector<Ant> &ants);

    /**
     * Macro-mutation: pick a random sub-block of the chain and re-fold it with random directions
     * while ensuring feasibility. If energy improves, keep it; otherwise revert.
     */
    bool macroMutation(Ant &ant);

    /**
     * 1-exchange: for each fold position, try each possible new direction,
     * see if it’s feasible and if it improves energy. If so, accept it.
     */
    bool oneExchange(Ant &ant);

    /**
     * Update global pheromone:
     *  - Evaporation: tau_{i,d} = (1-rho) * tau_{i,d}
     *  - Reinforcement: for each selected ant (best or all),
     *    tau_{i,d} += \Delta_{i,d,c} (like E(c)/E^*).
     *  - Then re-normalize if ratio(max/min) < stagnationThreshold.
     */
    void updatePheromones(const std::vector<Ant> &ants);

    /**
     * Evaluate the ant (energy = number of H-H contacts). Store in ant.score.
     */
    void evaluateAnt(Ant &ant);

    /**
     * Probability rule to choose direction at fold i.
     */
    DIRECTION chooseDirection(int foldIndex, Ant &ant);

    /**
     * Utility: random in [0,1).
     */
    double rand01() {
        return std::uniform_real_distribution(0.0, 1.0)(rng);
    }

    bool attemptFold(int index, Ant &ant, bool forward);

    int calculateNewContacts(int foldIndex, int direction, const Ant &ant);

    bool longRangeMutation(Ant &ant, int mutationPoint);

    bool isDirectionFeasible(int foldIndex, DIRECTION dir, const HPModel &model);

    int foldIndex(int fold, int currentFolds);

    void buildChain(Ant &ant, int leftPos, int rightPos);

    bool performLongRangeMutation(Ant &ant);
    std::string convertFoldsToString(const std::vector<DIRECTION> &folds);

};

