//
// Created by William on 23/10/2024.
//

#include <utility>
#include <vector>
#include "FileCacheManager.h"
#include <fstream>
#include <iostream>
#include <filesystem>
#include <algorithm>
#include <map>

// Note that this path is almost certainly going to break if program is precompiled
const std::string DATA_PREFIX = "../resources/data/";
const std::string PROGRESS_PREFIX = "inProgress/";
const std::string INDEX_FILE = "index.csv";
const std::string TODO_FILE = "todo.txt";
const int SEQUENCE_INDEX = 0;
const int OPTIMAL_SCORE_INDEX = 1;
const int OPTIMAL_FOLD_INDEX = 2;
const int STATUS_INDEX = 3;
const int STRATEGY_INDEX = 4;
const int DATA_LINK_INDEX = 5;
const std::map<std::string, StrategyType> STRATEGY_TABLE = {
        {"BRUTE_FORCE", BRUTE_FORCE}
};
const std::map<std::string, SequenceStatus> SEQUENCE_TABLE = {
        {"OPTIMAL_FOUND", OPTIMAL_FOUND},
        {"IN_PROGRESS",   IN_PROGRESS},
        {"NOT_FOUND",     NOT_FOUND}
};

std::string random_string(size_t length) {
    auto randchar = []() -> char {
        const char charset[] =
                "0123456789"
                "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
                "abcdefghijklmnopqrstuvwxyz";
        const size_t max_index = (sizeof(charset) - 1);
        return charset[rand() % max_index];
    };
    std::string str(length, 0);
    std::generate_n(str.begin(), length, randchar);
    return str;
}

std::string readDataFile(const std::string& dataFileName) {

}

std::vector<std::string> readFile(const std::string& fileName) {
    std::string path = DATA_PREFIX + fileName;
    std::ifstream file(path);

    std::string rawText;
    std::vector<std::string> result;
    while (std::getline(file, rawText)) {
        result.push_back(rawText);
    }
    return result;
}

void deleteToDoIfExists(const std::string &sequenceToDelete) {
    auto todos = getSequencesToDo();
    std::string newTodos;
    bool matchFound = false;
    for (auto todo: todos) {
        if (todo != sequenceToDelete) {
            newTodos += todo;
        } else {
            matchFound = true;
            break;
        }
    }
    // If no match is found, we can just leave the list as it is
    if (matchFound) {
        std::string path = DATA_PREFIX + TODO_FILE;
        std::ofstream output(path);
        output << newTodos;
    }

}

void writeToIndex(const std::string &sequence, SequenceStatus status, StrategyType strategy, const std::string &folds,
                  int score, const std::string &dataLink) {
    auto lines = readFile(INDEX_FILE);
    int index = -1;
    for (int i = 0; i < lines.size(); i++) {
        if (lines[i].contains(sequence)) {
            index = i;
            break;
        }
    }
    std::string newLine = sequence + "," + std::to_string(score) + "," + folds + "," + SequenceNames[status] + "," +
                          StrategyNames[strategy] + "," + dataLink;
    if (index == -1) {
        lines.push_back(newLine);
    } else {
        lines[index] = newLine;
    }

    std::ofstream output(DATA_PREFIX + INDEX_FILE);
    std::string rawString;
    for (const auto &l: lines) {
        rawString += l;
        rawString += "\n";
    }
    output << rawString;
    if (status == OPTIMAL_FOUND) {
        deleteToDoIfExists(sequence);
    }
}

void writeToProgressFile(const std::string &name, LongTermCacheData data) {
    std::string path = DATA_PREFIX + PROGRESS_PREFIX + name;
    std::ofstream output(path);
    output << data.toString();
}

void createProgressFile(const std::string &name) {
    std::string path = DATA_PREFIX + PROGRESS_PREFIX + name;
    std::ofstream output(path);
}


std::vector<std::string> splitCSVString(const std::string &str) {
    std::stringstream ss(str);

    std::string t;

    char del = ',';

    std::vector<std::string> res;

    while (std::getline(ss, t, del)) {
        res.push_back(t);
    }
    return res;
}

std::vector<std::string> readCSVLine(const std::string &sequence, StrategyType strategyType) {
    auto lines = readFile(INDEX_FILE);
    std::string foundSequence;
    std::vector<std::string> res;
    std::string strategyString = StrategyNames[strategyType];
    for (const auto &line: lines) {
        if (line.contains(sequence) && line.contains(strategyString)) {
            foundSequence = line;
            break;
        }
    }
    if (foundSequence.empty()) {
        return res;
    }

    return splitCSVString(foundSequence);

}

SequenceStatus getSequenceStatus(const std::string &sequence, StrategyType strategy) {
    auto line = readCSVLine(sequence, strategy);
    if (line.size() == 0) {
        return NOT_FOUND;
    }
    std::string statusString = line[STATUS_INDEX];
    return SEQUENCE_TABLE.at(statusString);
}

void saveSequenceInProgress(const std::string &sequence, StrategyType strategyType, int score, LongTermCacheData data) {
    auto line = readCSVLine(sequence, strategyType);
    std::string id;
    if (line.empty()) {
        // No previous data exists
        std::string newId = random_string(32);
        createProgressFile(newId);
        id = newId;
    } else {
        id = line[DATA_LINK_INDEX];
        if (id == "NONE") {
            id = random_string(32);
        }
    }

    writeToProgressFile(id, std::move(data));
    writeToIndex(sequence, IN_PROGRESS, strategyType, "", score, id);
}

void deleteProgressFile(const std::string &name) {
    std::string path = DATA_PREFIX + PROGRESS_PREFIX + name;
    std::filesystem::remove(path);
}


void
saveSequence(const std::string &sequence, SequenceStatus status, StrategyType strategyType, const std::string &folds,
             int score) {
    if (status == IN_PROGRESS) {
        throw std::invalid_argument("Can not save sequence in progress without LongTermCacheData defined\n");
    }
    // We need to check if the sequence is currently in progress, so we can clean up before finalizing the sequence
    // This avoids leaking files in the data folder
    auto existingStatus = getSequenceStatus(sequence, strategyType);
    if (existingStatus == IN_PROGRESS) {
        auto lines = readCSVLine(sequence, strategyType);
        std::string dataLink = lines[DATA_LINK_INDEX];
        deleteProgressFile(dataLink);
    }

    writeToIndex(sequence, status, strategyType, folds, score, "NONE");
}

void saveSequence(const std::string &sequence, StrategyType strategyType, int score, LongTermCacheData data) {
    saveSequenceInProgress(sequence, strategyType, score, std::move(data));
}

std::vector<std::string> getSequencesToDo() {
    return readFile(TODO_FILE);
}

std::string getDataFileName(const std::string& sequence, StrategyType type) {
    auto indexLines = readFile(INDEX_FILE);
    for (auto line : indexLines) {
        if (line.contains(sequence) && line.contains(StrategyNames[type])) {
            auto split = splitCSVString(line);
            std::string link = split[DATA_LINK_INDEX];
            if (link != "NONE") {
                return link;
            }
        }
    }
    return "";
}

LongTermCacheData loadSequence(const std::string &sequence, StrategyType type) {
    LongTermCacheData data = LongTermCacheData{};
    std::string dataLink = getDataFileName(sequence, type);
    if (dataLink.empty()) {
        return data;
    }
    std::string dataString = readDataFile(dataLink);
    data.fromDataString(dataString);
    return data;
}
